﻿using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.Linq;
using System.Linq.Expressions;
using System.Threading.Tasks;
using ClientServer.Controllers.Core.Users;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models.Exercises;
using ClientServer.Models.Exercises.Release;
using ClientServer.Models.Exercises.Solution;
using ClientServer.Models.Users;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Controllers.Core.Exercises
{
    /// <summary>
    /// controller for the exercise overview page
    /// </summary>
    [Route(Constants.ApiPrefix + "exercise/overview")]
    public class ExerciseOverviewController : ControllerWithDb
    {
        public static List<string> possibleLogicalConnectors = new List<string>() {"and", "or"};

        public ExerciseOverviewController(YapexDbContext context) : base(context)
        {
        }

        //old
//        [HttpGet("all")]
//        public async void GetAccessibleExercises()
//        {
//            if (!await base.IsLoggedIn()) return;
//
//            int userId = GetUserId();
//
//            // get all
//
//            //get the participations where the user has a solution (participation exists)
//            //when entering a release code the user gets automatically a participation so this is included here
//            //when the release ended or the working time is up then the access will be readonly
//            List<ExerciseReleaseWithUserAsParticipation> releasedWithCodeOrHasSolution =
//                _context.ExerciseReleaseWithUserAsParticipations
//                    .Include(p => p.ExerciseRelease)
//                    .ThenInclude(release => release.Exercise)
//                    .ThenInclude(exercise => exercise.MetaData)
//                    .ThenInclude(metaData => metaData.TagWithMetaDatas)
//                    .ThenInclude(p => p.Tag)
//                    .Where(p =>
//                        p.UserId == userId &&
//                        p.ExerciseRelease.HideInOverviews == false
//                    )
//                    .ToList();
//
//            // exercises that are accessible explicite to all users (exercise property)
//
//            //this might return the same (or just some) exercises as the participation ... 
//            //so check if the user already have a participation for this --> then it's included in the previous query... else include this here
//            List<ExerciseRelease> releasedWithoutCode = _context.ExerciseReleases
//                .Include(p => p.Exercise)
//                .ThenInclude(exercise => exercise.MetaData)
//                .ThenInclude(metaData => metaData.TagWithMetaDatas)
//                .ThenInclude(p => p.Tag)
//                .Where(p => p.IsReleased && p.IsVisibleToAllAfterRelease
//                                         && p.ExerciseReleaseWithUserAsParticipations.Count(parti =>
//                                             parti.UserId == userId) == 0)
//                .ToList();
//
//            var exercisePreviewsForFrontend = new List<ExercisePreviewFromBackend>();
//
//            foreach (ExerciseReleaseWithUserAsParticipation access in releasedWithCodeOrHasSolution)
//            {
//                exercisePreviewsForFrontend.Add(new ExercisePreviewFromBackend()
//                {
//                    Id = access.ExerciseRelease.ExerciseId,
//                    DisplayName = access.ExerciseRelease.Exercise.DisplayName,
//                    ShortDescription = access.ExerciseRelease.Exercise.ShortDescription,
//                    GeneratedCode = access.ExerciseRelease.GeneratedCode,
//                    ShouldNotCount = access.ShouldNotCount,
//                    Tags = access.ExerciseRelease.Exercise.MetaData.TagWithMetaDatas.Select(p => new TagFullBase()
//                    {
//                        Id = p.Tag.Id,
//                        DisplayName = p.Tag.DisplayName,
//                        HtmlIcon = p.Tag.HtmlIcon,
//                        HtmlColor = p.Tag.HtmlColor,
//                        Description = p.Tag.Description,
//                        HtmlBackgroundColor = p.Tag.HtmlBackgroundColor,
//                    }).ToList(),
//                    ReleasedForPLangId = access.ExerciseRelease.PLangId,
//                    IsVisibleToAllAfterRelease = access.ExerciseRelease.IsVisibleToAllAfterRelease,
//                    ReleaseDurationType = (int) access.ExerciseRelease.ReleaseDurationType,
//                    ReleaseStartType = (int) access.ExerciseRelease.ReleaseStartType,
//                    HasLimitedWorkingTime = access.ExerciseRelease.HasLimitedWorkingTime,
//                    AvailableWorkingTimeInMinutes = access.ExerciseRelease.AvailableWorkingTimeInMinutes,
//                    AutomaticEndAt = access.ExerciseRelease.AutomaticEndAt,
//                    FirstAccessedAt = access.CreatedAt,
//                    IsReleased = access.ExerciseRelease.IsReleased,
//                });
//            }
//
//            foreach (var exerciseRelease in releasedWithoutCode)
//            {
//                exercisePreviewsForFrontend.Add(new ExercisePreviewFromBackend()
//                {
//                    Id = exerciseRelease.ExerciseId,
//                    DisplayName = exerciseRelease.Exercise.DisplayName,
//                    ShortDescription = exerciseRelease.Exercise.ShortDescription,
//                    GeneratedCode = exerciseRelease.GeneratedCode,
//                    Tags = exerciseRelease.Exercise.MetaData.TagWithMetaDatas.Select(p => new TagFullBase()
//                    {
//                        Id = p.Tag.Id,
//                        DisplayName = p.Tag.DisplayName,
//                        HtmlIcon = p.Tag.HtmlIcon,
//                        HtmlColor = p.Tag.HtmlColor,
//                        Description = p.Tag.Description,
//                        HtmlBackgroundColor = p.Tag.HtmlBackgroundColor,
//                    }).ToList(),
//                    ReleasedForPLangId = exerciseRelease.PLangId,
//                    IsVisibleToAllAfterRelease = exerciseRelease.IsVisibleToAllAfterRelease,
//                    ReleaseDurationType = (int) exerciseRelease.ReleaseDurationType,
//                    ReleaseStartType = (int) exerciseRelease.ReleaseStartType,
//                    HasLimitedWorkingTime = exerciseRelease.HasLimitedWorkingTime,
//                    AvailableWorkingTimeInMinutes = exerciseRelease.AvailableWorkingTimeInMinutes,
//                    AutomaticEndAt = exerciseRelease.AutomaticEndAt,
//                    FirstAccessedAt = null,
//                    IsReleased = exerciseRelease.IsReleased,
//                });
//            }
//
//
//            await
//                Response.WriteAsync(
//                    Jc.Serialize(new BasicResponseWithData<List<ExercisePreviewFromBackend>>(ResponseCode.Ok,
//                        "all accessible exercises queried",
//                        exercisePreviewsForFrontend)));
//        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="pagination"></param>
        /// <param name="viaCode">true: accessible via code, false: through is visible for all</param>
        [HttpPost("all/closed/{viaCode}")]
        public async Task GetAccessibleClosedExercises([FromBody] PaginationPostDataWithTagsFilter pagination,
            bool viaCode)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            IQueryable<ExerciseReleaseWithUserAsParticipation> participations = null;

            if (viaCode)
            {
                //release must be closed, (released with code or participation)
                // and not hidden

                participations =
                    _context.ExerciseReleaseWithUserAsParticipations
                        .Include(p => p.ExerciseRelease)
                        .ThenInclude(release => release.Exercise)
                        .ThenInclude(exercise => exercise.MetaData)
                        .ThenInclude(metaData => metaData.TagWithMetaDatas)
                        .ThenInclude(p => p.Tag)
                        .Where(p =>
                            p.UserId == userId
                            && p.ExerciseRelease.IsReleased == false
                            && p.ExerciseRelease.HideInOverviews == false
                            && p.ExerciseRelease.IsVisibleToAllAfterRelease == false
                        )
                    ;
                //when we have multiple plangs per participation this is not the same as getting the participation 
                //but currently releases can only have 1 plang
//                releases =
//                    _context.ExerciseReleases
//                        .Include(p => p.Exercise)
//                        .ThenInclude(exercise => exercise.MetaData)
//                        .ThenInclude(metaData => metaData.TagWithMetaDatas)
//                        .ThenInclude(p => p.Tag)
//                        .Where(p =>
//                            p.IsReleased == false
//                            && p.IsVisibleToAllAfterRelease == false
//                            && p.HideInOverviews == false
//                            && p.ExerciseReleaseWithUserAsParticipations.Count(part =>
//                                part.UserId == userId) > 0
//                        )
//                    ;
            }
            else
            {
                //via is visible for all
                //release must be closed, visible for all and the user needs to have a solution (participation)
                // and not hidden
//                releases = _context.ExerciseReleases
//                        .Include(p => p.Exercise)
//                        .ThenInclude(exercise => exercise.MetaData)
//                        .ThenInclude(metaData => metaData.TagWithMetaDatas)
//                        .ThenInclude(p => p.Tag)
//                        .Where(p => p.IsReleased == false && p.IsVisibleToAllAfterRelease
//                                                          && p.ExerciseReleaseWithUserAsParticipations.Count(part =>
//                                                              part.UserId == userId) > 0
//                                                          && p.HideInOverviews == false
//                        )
//                    ;
                //but currently releases can only have 1 plang
                participations =
                    _context.ExerciseReleaseWithUserAsParticipations
                        .Include(p => p.ExerciseRelease)
                        .ThenInclude(release => release.Exercise)
                        .ThenInclude(exercise => exercise.MetaData)
                        .ThenInclude(metaData => metaData.TagWithMetaDatas)
                        .ThenInclude(p => p.Tag)
                        .Where(p =>
                            p.UserId == userId
                            && p.ExerciseRelease.IsReleased == false
                            && p.ExerciseRelease.HideInOverviews == false
                            && p.ExerciseRelease.IsVisibleToAllAfterRelease
                        )
                    ;
            }

            participations = _applyTagFilters(participations, pagination);

            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                participations = participations
                    .Where(p => (p.ExerciseRelease.Exercise.DisplayName != null &&
                                 p.ExerciseRelease.Exercise.DisplayName.ToLower().Contains(text))
                                || (p.ExerciseRelease.Exercise.ShortDescription != null &&
                                    p.ExerciseRelease.Exercise.ShortDescription.ToLower().Contains(text))
                    );
            }

            string orderByProp = "";
            Expression<Func<ExerciseReleaseWithUserAsParticipation, object>> orderByFunc = null;

            //not all props are sortable via the frontend
            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "Id".ToLower())
                {
                    orderByFunc = p => p.ExerciseRelease.Exercise.Id;
                    orderByProp = "Id".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "DisplayName".ToLower())
                {
                    orderByFunc = p => p.ExerciseRelease.Exercise.DisplayName;
                    orderByProp = "DisplayName".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "ShortDescription".ToLower())
                {
                    orderByFunc = p => p.ExerciseRelease.Exercise.ShortDescription;
                    orderByProp = "ShortDescription".ToLower();
                }
                //TODO not working for some reason... we get pgsql exception: ... missing in from clause
                //if we display the sql with var qqq = IQueryableExtensions.ToSql<T>(pagedQuery);
                //we get a valid statement... probably ToSql helper is not working 100%
//                else if (pagination.OrderByProp.ToLower() == "IsPermanentlyLocked".ToLower())
//                {
//                    orderByFunc = p => p.ExerciseRelease.Exercise.IsPermanentlyLocked;
//                    orderByProp = "IsPermanentlyLocked".ToLower();
//                }
                else
                {
                    orderByFunc = p => p.ExerciseRelease.Exercise.Id;
                    orderByProp = "Id".ToLower();
                }
            }

            var pagedData = await PaginationHelper.GetAsPaginationAsync<ExerciseReleaseWithUserAsParticipation>(
                participations,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );


            var temp = pagedData.Items.Select(participation => new ExercisePreviewFromBackend()
                {
                    Id = participation.ExerciseRelease.ExerciseId,
                    DisplayName = participation.ExerciseRelease.Exercise.DisplayName,
                    ShortDescription = participation.ExerciseRelease.Exercise.ShortDescription,
                    GeneratedCode = participation.ExerciseRelease.GeneratedCode,
                    Tags = participation.ExerciseRelease.Exercise.MetaData.TagWithMetaDatas.Select(p =>
                        new TagFullBase()
                        {
                            Id = p.Tag.Id,
                            DisplayName = p.Tag.DisplayName,
                            HtmlIcon = p.Tag.HtmlIcon,
                            HtmlColor = p.Tag.HtmlColor,
                            Description = p.Tag.Description,
                            HtmlBackgroundColor = p.Tag.HtmlBackgroundColor,
                        }).ToList(),
                    ReleasedForPLangId = participation.ExerciseRelease.PLangId,
                    IsPermanentlyLocked = participation.ExerciseRelease.Exercise.IsPermanentlyLocked,
                    ShouldNotCount = participation.ShouldNotCount,
                    IsVisibleToAllAfterRelease = participation.ExerciseRelease.IsVisibleToAllAfterRelease,
                    ReleaseDurationType = (int) participation.ExerciseRelease.ReleaseDurationType,
                    ReleaseStartType = (int) participation.ExerciseRelease.ReleaseStartType,
                    HasLimitedWorkingTime = participation.ExerciseRelease.HasLimitedWorkingTime,
                    AvailableWorkingTimeInMinutes = participation.ExerciseRelease.AvailableWorkingTimeInMinutes,
                    AutomaticEndAt = participation.ExerciseRelease.AutomaticEndAt,
                    FirstAccessedAt = null, //not needed because exercises are closed
                    IsReleased = participation.ExerciseRelease.IsReleased,
                })
                .ToList();

            var frontendData = PaginationHelper.TransformForFrontend(pagedData, temp);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<ExercisePreviewFromBackend>>(ResponseCode.Ok,
                        "all accessible exercises queried",
                        frontendData)));
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="pagination"></param>
        /// <param name="viaCode">true: accessible via code, false: through is visible for all</param>
        [HttpPost("all/open/{viaCode}")]
        public async Task GetAccessibleOpenExercises([FromBody] PaginationPostDataWithTagsFilter pagination,
            bool viaCode)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            PaginatedData<ExercisePreviewFromBackend> frontendData;
            var list = new List<ExercisePreviewFromBackend>();

            if (viaCode)
            {
                //release must be closed, (released with code or participation)
                // and not hidden
                var participations =
                        _context.ExerciseReleaseWithUserAsParticipations
                            .Include(p => p.ExerciseRelease)
                            .ThenInclude(release => release.Exercise)
                            .ThenInclude(exercise => exercise.MetaData)
                            .ThenInclude(metaData => metaData.TagWithMetaDatas)
                            .ThenInclude(p => p.Tag)
                            .Where(p =>
                                p.UserId == userId
                                && p.ExerciseRelease.IsReleased
                                && p.ExerciseRelease.IsVisibleToAllAfterRelease == false
                                && p.ExerciseRelease.HideInOverviews == false
                            )
                    ;

                if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
                {
                    var text = pagination.SearchText.ToLower();
                    participations = participations
                        .Where(p => (p.ExerciseRelease.Exercise.DisplayName != null &&
                                     p.ExerciseRelease.Exercise.DisplayName.ToLower().Contains(text))
                                    || (p.ExerciseRelease.Exercise.ShortDescription != null &&
                                        p.ExerciseRelease.Exercise.ShortDescription.ToLower().Contains(text))
                        );
                }

                participations = _applyTagFilters(participations, pagination);


                string orderByProp = "";
                Expression<Func<ExerciseReleaseWithUserAsParticipation, object>> orderByFunc = null;

                //not all props are sortable via the frontend
                if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
                {
                    if (pagination.OrderByProp.ToLower() == "Id".ToLower())
                    {
                        orderByFunc = p => p.ExerciseRelease.Exercise.Id;
                        orderByProp = "Id".ToLower();
                    }
                    else if (pagination.OrderByProp.ToLower() == "DisplayName".ToLower())
                    {
                        orderByFunc = p => p.ExerciseRelease.Exercise.DisplayName;
                        orderByProp = "DisplayName".ToLower();
                    }
                    else if (pagination.OrderByProp.ToLower() == "ShortDescription".ToLower())
                    {
                        orderByFunc = p => p.ExerciseRelease.Exercise.ShortDescription;
                        orderByProp = "ShortDescription".ToLower();
                    }
                    //TODO not working for some reason... we get pgsql exception: ... missing in from clause
//                    else if (pagination.OrderByProp.ToLower() == "IsPermanentlyLocked".ToLower())
//                    {
//                        orderByFunc = p => p.ExerciseRelease.Exercise.IsPermanentlyLocked;
//                        orderByProp = "IsPermanentlyLocked".ToLower();
//                    }
                    else
                    {
                        orderByFunc = p => p.ExerciseRelease.Exercise.Id;
                        orderByProp = "Id".ToLower();
                    }
                }

                var pagedData = await PaginationHelper.GetAsPaginationAsync(participations,
                    pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                    orderByFunc, orderByProp, pagination.SearchText
                );

                list = pagedData.Items.Select(connection => new ExercisePreviewFromBackend()
                    {
                        Id = connection.ExerciseRelease.Exercise.Id,
                        DisplayName = connection.ExerciseRelease.Exercise.DisplayName,
                        ShortDescription = connection.ExerciseRelease.Exercise.ShortDescription,
                        GeneratedCode = connection.ExerciseRelease.GeneratedCode,
                        Tags = connection.ExerciseRelease.Exercise.MetaData.TagWithMetaDatas.Select(p =>
                            new TagFullBase()
                            {
                                Id = p.Tag.Id,
                                DisplayName = p.Tag.DisplayName,
                                HtmlIcon = p.Tag.HtmlIcon,
                                HtmlColor = p.Tag.HtmlColor,
                                Description = p.Tag.Description,
                                HtmlBackgroundColor = p.Tag.HtmlBackgroundColor,
                            }).ToList(),
                        ReleasedForPLangId = connection.ExerciseRelease.PLangId,
                        IsVisibleToAllAfterRelease = connection.ExerciseRelease.IsVisibleToAllAfterRelease,
                        IsPermanentlyLocked = connection.ExerciseRelease.Exercise.IsPermanentlyLocked,
                        ShouldNotCount = connection.ShouldNotCount,
                        ReleaseDurationType = (int) connection.ExerciseRelease.ReleaseDurationType,
                        ReleaseStartType = (int) connection.ExerciseRelease.ReleaseStartType,
                        HasLimitedWorkingTime = connection.ExerciseRelease.HasLimitedWorkingTime,
                        AvailableWorkingTimeInMinutes = connection.ExerciseRelease.AvailableWorkingTimeInMinutes,
                        AutomaticEndAt = connection.ExerciseRelease.AutomaticEndAt,
                        FirstAccessedAt = connection.CreatedAt,
                        IsReleased = connection.ExerciseRelease.IsReleased,
                    })
                    .ToList();

                frontendData = PaginationHelper.TransformForFrontend(pagedData, list);
            }
            else
            {
                //via is visible for all
                //release must be closed, visible for all and the user needs to have a solution (participation)
                // and not hidden
                var releases = _context.ExerciseReleases
                        .Include(p => p.Exercise)
                        .ThenInclude(exercise => exercise.MetaData)
                        .ThenInclude(metaData => metaData.TagWithMetaDatas)
                        .ThenInclude(p => p.Tag)
                        .Where(p => p.IsReleased
                                    && p.IsVisibleToAllAfterRelease
                                    && p.HideInOverviews == false
                        )
                    ;

                releases = _applyTagFilters(releases, pagination);

                //TODO this is not good for pagination... but don't know how to do this better yet
                var releasesParticipations =
                        _context.ExerciseReleaseWithUserAsParticipations
                            .Where(p =>
                                p.UserId == userId
                                && p.ExerciseRelease.IsReleased == true
                                && p.ExerciseRelease.IsVisibleToAllAfterRelease
                                && p.ExerciseRelease.HideInOverviews == false
                            )
                            .ToList()
                    ;

                if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
                {
                    var text = pagination.SearchText.ToLower();
                    releases = releases
                        .Where(p => (p.Exercise.DisplayName != null && p.Exercise.DisplayName.ToLower().Contains(text))
                                    || (p.Exercise.ShortDescription != null &&
                                        p.Exercise.ShortDescription.ToLower().Contains(text))
                        );
                }

                string orderByProp = "";
                Expression<Func<ExerciseRelease, object>> orderByFunc = null;

                //not all props are sortable via the frontend
                if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
                {
                    if (pagination.OrderByProp.ToLower() == "id".ToLower())
                    {
                        orderByFunc = p => p.Id;
                        orderByProp = "id".ToLower();
                    }
                    else if (pagination.OrderByProp.ToLower() == "DisplayName".ToLower())
                    {
                        orderByFunc = p => p.Exercise.DisplayName;
                        orderByProp = "DisplayName".ToLower();
                    }
                    else if (pagination.OrderByProp.ToLower() == "ShortDescription".ToLower())
                    {
                        orderByFunc = p => p.Exercise.ShortDescription;
                        orderByProp = "ShortDescription".ToLower();
                    }
                    //TODO not working for some reason... we get pgsql exception: ... missing in from clause
//                    else if (pagination.OrderByProp.ToLower() == "IsPermanentlyLocked".ToLower())
//                    {
//                        orderByFunc = p => p.Exercise.IsPermanentlyLocked;
//                        orderByProp = "IsPermanentlyLocked".ToLower();
//                    }
                    else
                    {
                        orderByFunc = p => p.Id;
                        orderByProp = "id".ToLower();
                    }
                }

                var pagedData = await PaginationHelper.GetAsPaginationAsync(releases,
                    pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                    orderByFunc, orderByProp, pagination.SearchText
                );

                list = pagedData.Items.Select(exerciseRelease =>
                    {
                        //try to get a participation for the user for the release...
                        //if he didn't click the exercise yet there is no participation
                        var pariticpation =
                            releasesParticipations.FirstOrDefault(p => p.ExerciseReleaseId == exerciseRelease.Id);

                        return new ExercisePreviewFromBackend()
                        {
                            Id = exerciseRelease.ExerciseId,
                            DisplayName = exerciseRelease.Exercise.DisplayName,
                            ShortDescription = exerciseRelease.Exercise.ShortDescription,
                            GeneratedCode = exerciseRelease.GeneratedCode,
                            Tags = exerciseRelease.Exercise.MetaData.TagWithMetaDatas.Select(p => new TagFullBase()
                            {
                                Id = p.Tag.Id,
                                DisplayName = p.Tag.DisplayName,
                                HtmlIcon = p.Tag.HtmlIcon,
                                HtmlColor = p.Tag.HtmlColor,
                                Description = p.Tag.Description,
                                HtmlBackgroundColor = p.Tag.HtmlBackgroundColor,
                            }).ToList(),
                            ReleasedForPLangId = exerciseRelease.PLangId,
                            IsVisibleToAllAfterRelease = exerciseRelease.IsVisibleToAllAfterRelease,
                            IsPermanentlyLocked = exerciseRelease.Exercise.IsPermanentlyLocked,
                            ShouldNotCount =
                                false, //we would need another query for that... just hide in frontend for now
                            ReleaseDurationType = (int) exerciseRelease.ReleaseDurationType,
                            ReleaseStartType = (int) exerciseRelease.ReleaseStartType,
                            HasLimitedWorkingTime = exerciseRelease.HasLimitedWorkingTime,
                            AvailableWorkingTimeInMinutes = exerciseRelease.AvailableWorkingTimeInMinutes,
                            AutomaticEndAt = exerciseRelease.AutomaticEndAt,
                            FirstAccessedAt = pariticpation?.CreatedAt,
                            IsReleased = exerciseRelease.IsReleased,
                        };
                    })
                    .ToList();

                frontendData = PaginationHelper.TransformForFrontend(pagedData, list);
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<ExercisePreviewFromBackend>>(ResponseCode.Ok,
                        "all accessible exercises queried",
                        frontendData)));
        }

        /// <summary>
        /// returns the groups where the user has permission to do something with the groups exercises
        /// almost the same as getting a single group but here the user also needs permission in the group 
        /// </summary>
        [HttpGet("groups")]
        public async Task GetGroupsForExercises()
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            //only get exercises where the user has permission
            List<UserWithUserGroup> connections = await _context.UserWithUserGroups
                .Where(
                    p => p.UserId == userId
                         && (p.GroupRole.GroupRolePermission.CanCreateExercises
                             || p.GroupRole.GroupRolePermission.CanChangeExercises
                             || p.GroupRole.GroupRolePermission.CanDeleteExercises
                             || p.GroupRole.GroupRolePermission.CanAssessExercises
                             || p.GroupRole.GroupRolePermission.CanManageExerciseReleases
                             || p.GroupRole.GroupRolePermission.CanSeeExercisesFromOthersInGroup
                         )
                )
                .Include(p => p.UserGroup)
                .Include(p => p.GroupRole)
                .Include(p => p.GroupRole.GroupRolePermission)
                .ToListAsync();

            var frontendData = connections.Select(p => new UserGroupFromBackend()
                {
                    Id = p.UserGroupId,
                    Email = p.UserGroup.Email,
                    DisplayName = p.UserGroup.DisplayName,
                    GroupRole = new GroupRoleWithPermissionFullBase
                    (
                        id: p.GroupRole.Id,
                        displayName: p.GroupRole.DisplayName,
                        canChangeExercises: p.GroupRole.GroupRolePermission.CanChangeExercises,
                        canCreateExercises: p.GroupRole.GroupRolePermission.CanCreateExercises,
                        canAssessExercises: p.GroupRole.GroupRolePermission.CanAssessExercises,
                        canManageExerciseReleases: p.GroupRole.GroupRolePermission.CanManageExerciseReleases,
                        canDeleteExercises: p.GroupRole.GroupRolePermission.CanDeleteExercises,
                        canAddUserToGroup: p.GroupRole.GroupRolePermission.CanAddUserToGroup,
                        canLockExercisesPermanently: p.GroupRole.GroupRolePermission.CanLockExercisesPermanently,
                        canSeeExercisesFromOthersInGroup: p.GroupRole.GroupRolePermission
                            .CanSeeExercisesFromOthersInGroup,
                        canChangeOtherMembersRole:
                        p.GroupRole.GroupRolePermission.CanChangeOtherMembersRole,
                        canRemoveMemberFromGroup:
                        p.GroupRole.GroupRolePermission.CanRemoveMemberFromGroup,
                        canSeeOtherMembers: p.GroupRole.GroupRolePermission.CanSeeOtherMembers,
                        canChangeGroupData: p.GroupRole.GroupRolePermission.CanChangeGroupData
                    )
                })
                .ToList();

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<UserGroupFromBackend>>(ResponseCode.Ok,
                        "",
                        frontendData)));
        }

        /// <summary>
        /// returns the exercises (only description) for the given group (the user needs permission)
        /// we don't include groups where the user hav no permission but own some exercise... use the own exercises overview
        /// </summary>
        [HttpPost("groups/{groupId}/exercises")]
        public async Task GetAccessibleExercisesThroughGroups([FromBody] PaginationPostDataWithTagsFilter pagination,
            int groupId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            var connection = await _context.UserWithUserGroups
                    .FirstOrDefaultAsync(p => p.UserId == userId && p.UserGroupId == groupId
                                                                 && (p.GroupRole.GroupRolePermission.CanCreateExercises
                                                                     || p.GroupRole.GroupRolePermission
                                                                         .CanChangeExercises
                                                                     || p.GroupRole.GroupRolePermission
                                                                         .CanDeleteExercises
                                                                     || p.GroupRole.GroupRolePermission
                                                                         .CanAssessExercises
                                                                     || p.GroupRole.GroupRolePermission
                                                                         .CanManageExerciseReleases
                                                                     || p.GroupRole.GroupRolePermission
                                                                         .CanSeeExercisesFromOthersInGroup
                                                                 ))
                ;

            if (connection == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "group not found or no permission")));
                return;
            }

            var temp = _context.Exercises
                    .Include(p => p.MetaData)
                    .Include(p => p.MetaData.TagWithMetaDatas)
                    .Include(p => p.User)
                    .Where(p => p.UserGroupId == groupId && (!p.IsOnlyVisibleToOwner || p.UserId == userId))
                ;

            temp = _applyTagFilters(temp, pagination);

            var exercises = temp
                    .Select(ex =>
                        new
                            EditableExercisePreviewFromBackend //TODO performance... before we done this after pagination but sorting was not working... because of the tags sub query somehow
                            {
                                Id = ex.Id,
                                DisplayName = ex.DisplayName,
                                ShortDescription = ex.ShortDescription,
                                Tags = ex.MetaData.TagWithMetaDatas.Select(p => new TagFullBase()
                                {
                                    Id = p.Tag.Id,
                                    DisplayName = p.Tag.DisplayName,
                                    HtmlIcon = p.Tag.HtmlIcon,
                                    HtmlColor = p.Tag.HtmlColor,
                                    Description = p.Tag.Description,
                                    HtmlBackgroundColor = p.Tag.HtmlBackgroundColor,
                                }).ToList(),
                                IsOnlyVisibleToMe = ex.IsOnlyVisibleToOwner,
                                IsPermanentlyLocked = ex.IsPermanentlyLocked,
                                LastUpdatedAt = ex.LastUpdatedAt,
                                //Solutions = ex.Releases.Select(p => p.ExerciseReleaseWithUserAsParticipations).Count() //4 queries
                                Solutions = ex.Releases.Sum(p => p.ExerciseReleaseWithUserAsParticipations.Count),
                                Releases = ex.Releases.Count,
                                // 2 queries (aggregate function is ued)
                                ManagingGroupId = ex.UserGroupId,
                                SupportedPLangIds = ex.CodeTemplates.Select(p => p.PLangId).ToList(),
//                                tooooooo slow ... see workaround below
//                                Owner = ex.User != null
//                                    ? new ExerciseOwner(ex.User.Token, ex.User.FirstName, ex.User.LastName)
//                                    : null,
                                Owner = null,
                                OwnerId = ex.UserId
                            }
                    )
                ;


            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                exercises = exercises
                    .Where(p => (p.DisplayName != null && p.DisplayName.ToLower().Contains(text))
                                || (p.ShortDescription != null && p.ShortDescription.ToLower().Contains(text))
                    );
            }

            string orderByProp = "";
            Expression<Func<EditableExercisePreviewFromBackend, object>> orderByFunc = null;

            //not all props are sortable via the frontend
            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "id".ToLower())
                {
                    orderByFunc = p => p.Id;
                    orderByProp = "id".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "DisplayName".ToLower())
                {
                    orderByFunc = p => p.DisplayName;
                    orderByProp = "DisplayName".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "ShortDescription".ToLower())
                {
                    orderByFunc = p => p.ShortDescription;
                    orderByProp = "ShortDescription".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "IsOnlyVisibleToMe".ToLower())
                {
                    orderByFunc = p => p.IsOnlyVisibleToMe;
                    orderByProp = "IsOnlyVisibleToMe".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "LastUpdatedAt".ToLower())
                {
                    orderByFunc = p => p.LastUpdatedAt;
                    orderByProp = "LastUpdatedAt".ToLower();
                }

                else if (pagination.OrderByProp.ToLower() == "solutions".ToLower())
                {
                    orderByFunc = p => p.Solutions;
                    orderByProp = "solutions".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "releases".ToLower())
                {
                    orderByFunc = p => p.Releases;
                    orderByProp = "releases".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "IsPermanentlyLocked".ToLower())
                {
                    orderByFunc = p => p.IsPermanentlyLocked;
                    orderByProp = "IsPermanentlyLocked".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "OwnerId".ToLower())
                {
                    orderByFunc = p => p.OwnerId;
                    orderByProp = "OwnerId".ToLower();
                }
                else
                {
                    orderByFunc = p => p.Id;
                    orderByProp = "id".ToLower();
                }
            }

            PaginatedData<EditableExercisePreviewFromBackend> frontendData = await PaginationHelper.GetAsPaginationAsync(exercises,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );

            //TODO...
            //--- BIG workaround because querying the owner user is way too slow (group page load was 650ms with users 3000s)
            // other get exercise functions throw if we try to get the owner and exercise inline (in one select query)
            // probably so slow because ef tries to fetch the owner for each exercise separately??
            var ownerIds = frontendData.Items
                    .Select(p => p.OwnerId)
                    .ToImmutableHashSet()
                    .Where(p => p != null)
                    .ToList()
                ;

            //probably the db query will fail if we have a X long list via .Contains... so use a magic number here
            if (ownerIds.Count < Constants.MaxNumberOfOwnersToQueryInGroup)
            {
                var owners = await _context.Users
                        .Where(p => ownerIds.Contains(p.Id))
                        .ToDictionaryAsync(p => p.Id)
                    ;

                foreach (var item in frontendData.Items)
                {
                    if (item.OwnerId == null) continue;

                    User owner = null;

                    if (owners.TryGetValue(item.OwnerId.Value, out owner))
                    {
                        item.Owner = new ExerciseOwner(owner.Id, owner.Token, owner.FirstName, owner.LastName);
                    }
                }
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<EditableExercisePreviewFromBackend>>(
                        ResponseCode.Ok,
                        "",
                        frontendData)));
        }


        /// <summary>
        /// returns the exercises (only description) that are owned by the user
        /// </summary>
        [HttpPost("own")]
        public async Task GetOwnExercises([FromBody] PaginationPostDataWithTagsFilter pagination)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var owner = await _context.Users
                    .AsNoTracking()
                    .FirstOrDefaultAsync(p => p.Id == userId)
                ;

            var exOwner = new ExerciseOwner(owner.Id, owner.Token, owner.FirstName, owner.LastName);

            var temp = _context.Exercises
                    .Include(p => p.MetaData)
                    .Include(p => p.MetaData.TagWithMetaDatas)
                    .Include(p => p.User)
                    .Where(p => p.UserId == userId)
                    .AsNoTracking()
                ;

            temp = _applyTagFilters(temp, pagination);

            var userExercises = temp
                    .Select(ex =>
                        new
                            EditableExercisePreviewFromBackend //TODO performance... before we done this after pagination but sorting was not working... because of the tags sub query somehow
                            {
                                Id = ex.Id,
                                DisplayName = ex.DisplayName,
                                ShortDescription = ex.ShortDescription,
                                Tags = ex.MetaData.TagWithMetaDatas.Select(p => new TagFullBase()
                                {
                                    Id = p.Tag.Id,
                                    DisplayName = p.Tag.DisplayName,
                                    HtmlIcon = p.Tag.HtmlIcon,
                                    HtmlColor = p.Tag.HtmlColor,
                                    Description = p.Tag.Description,
                                    HtmlBackgroundColor = p.Tag.HtmlBackgroundColor,
                                }).ToList(),
                                IsOnlyVisibleToMe = ex.IsOnlyVisibleToOwner,
                                IsPermanentlyLocked = ex.IsPermanentlyLocked,
                                LastUpdatedAt = ex.LastUpdatedAt,
                                //Solutions = ex.Releases.Select(p => p.ExerciseReleaseWithUserAsParticipations).Count() //4 queries
                                Solutions = ex.Releases.Sum(p => p.ExerciseReleaseWithUserAsParticipations.Count),
                                Releases = ex.Releases.Count,
                                // 2 queries (aggregate function is ued)
                                ManagingGroupId = ex.UserGroupId,
                                SupportedPLangIds = ex.CodeTemplates.Select(p => p.PLangId).ToList(),

                                Owner = null, //the current user is the owner... only get the data once
                                OwnerId = ex.UserId
                            }
                    )
                ;


            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                userExercises = userExercises
                    .Where(p => (p.DisplayName != null && p.DisplayName.ToLower().Contains(text))
                                || (p.ShortDescription != null && p.ShortDescription.ToLower().Contains(text))
                    );
            }

            string orderByProp = "";
            Expression<Func<EditableExercisePreviewFromBackend, object>> orderByFunc = null;

            //not all props are sortable via the frontend
            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "id".ToLower())
                {
                    orderByFunc = p => p.Id;
                    orderByProp = "id".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "DisplayName".ToLower())
                {
                    orderByFunc = p => p.DisplayName;
                    orderByProp = "DisplayName".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "ShortDescription".ToLower())
                {
                    orderByFunc = p => p.ShortDescription;
                    orderByProp = "ShortDescription".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "IsOnlyVisibleToMe".ToLower())
                {
                    orderByFunc = p => p.IsOnlyVisibleToMe;
                    orderByProp = "IsOnlyVisibleToMe".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "LastUpdatedAt".ToLower())
                {
                    orderByFunc = p => p.LastUpdatedAt;
                    orderByProp = "LastUpdatedAt".ToLower();
                }

                else if (pagination.OrderByProp.ToLower() == "solutions".ToLower())
                {
                    orderByFunc = p => p.Solutions;
                    orderByProp = "solutions".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "releases".ToLower())
                {
                    orderByFunc = p => p.Releases;
                    orderByProp = "releases".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "IsPermanentlyLocked".ToLower())
                {
                    orderByFunc = p => p.IsPermanentlyLocked;
                    orderByProp = "IsPermanentlyLocked".ToLower();
                }
                else
                {
                    orderByFunc = p => p.Id;
                    orderByProp = "id".ToLower();
                }
            }


            var frontendData = await PaginationHelper.GetAsPaginationAsync(userExercises,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );

            foreach (var data in frontendData.Items)
            {
                data.Owner = exOwner;
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<EditableExercisePreviewFromBackend>>(
                        ResponseCode.Ok,
                        "own exercises queried",
                        frontendData)));
        }


        [HttpGet("own/exercise/{exerciseId}")]
        public async Task GetSingleOwnExercisesPreview(int exerciseId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            EditableExercisePreviewFromBackend userExercise;

            var owner = await _context.Users
                    .AsNoTracking()
                    .FirstOrDefaultAsync(p => p.Id == userId)
                ;


            userExercise = await _context.Exercises
                .Include(p => p.User)
                .Where(p => p.UserId == userId && p.Id == exerciseId)
                .Select(ex => new EditableExercisePreviewFromBackend
                {
                    Id = ex.Id,
                    DisplayName = ex.DisplayName,
                    ShortDescription = ex.ShortDescription,
                    Tags = ex.MetaData.TagWithMetaDatas.Select(p => new TagFullBase()
                    {
                        Id = p.Tag.Id,
                        DisplayName = p.Tag.DisplayName,
                        HtmlIcon = p.Tag.HtmlIcon,
                        HtmlColor = p.Tag.HtmlColor,
                        Description = p.Tag.Description,
                        HtmlBackgroundColor = p.Tag.HtmlBackgroundColor,
                    }).ToList(),
                    IsOnlyVisibleToMe = ex.IsOnlyVisibleToOwner,
                    IsPermanentlyLocked = ex.IsPermanentlyLocked,
                    LastUpdatedAt = ex.LastUpdatedAt,
                    //Solutions = ex.Releases.Select(p => p.ExerciseReleaseWithUserAsParticipations).Count() //4 queries
                    Solutions = ex.Releases.Sum(p => p.ExerciseReleaseWithUserAsParticipations.Count),
                    Releases = ex.Releases.Count,
                    // 2 queries (aggregate function is ued)
                    ManagingGroupId = ex.UserGroupId,
                    SupportedPLangIds = ex.CodeTemplates.Select(p => p.PLangId).ToList(),

                    Owner = null, //getting owner here results in an exception (Operation is already in progress)
                    OwnerId = ex.UserId
                })
                .AsNoTracking()
                .FirstOrDefaultAsync();

            userExercise.Owner = new ExerciseOwner(owner.Id, owner.Token, owner.FirstName, owner.LastName);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<EditableExercisePreviewFromBackend>(ResponseCode.Ok,
                        "own exercises queried",
                        userExercise)));
        }

        [HttpGet("group/exercise/{exerciseId}")]
        public async Task GetSingleGroupExercisesPreview(int exerciseId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            Exercise exercise = await _context.Exercises
                    .Include(p => p.UserGroup) //exercise could be hidden...
                    .Include(p => p.MetaData)
                    .Include(p => p.MetaData.TagWithMetaDatas)
                    .ThenInclude(p => p.Tag)
                    .Include(p => p.CodeTemplates)
                    .Include(p => p.User)
                    .AsNoTracking()
                    .FirstOrDefaultAsync(p => p.Id == exerciseId
                                              && (!p.IsOnlyVisibleToOwner || p.UserId == userId))
                ;

            if (exercise == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "group exercise not found")));
                return;
            }

            var userGroupId = exercise.UserGroup.Id;

            var managingGroupConnection = await _context.UserWithUserGroups
                    .Include(p => p.GroupRole)
                    .Include(p => p.GroupRole.GroupRolePermission)
                    .Include(p => p.UserGroup)
                    .AsNoTracking()
                    .FirstOrDefaultAsync(p => p.UserId == userId
                                              && p.UserGroupId == userGroupId)
                ;

            if (managingGroupConnection == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "group for exercise not found")));
                return;
            }

            var groupPermission = managingGroupConnection?.GroupRole?.GroupRolePermission;
            if (groupPermission == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no permission found for group")));
                return;
            }

            if (groupPermission.CanCreateExercises
                || groupPermission.CanChangeExercises
                || groupPermission.CanDeleteExercises
                || groupPermission.CanAssessExercises
                || groupPermission.CanManageExerciseReleases
                || groupPermission.CanSeeExercisesFromOthersInGroup
                || exercise.UserId == userId //or it's our own exercise
            )
            {
                //ok
            }
            else
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            var tagsForFrontend = new List<TagFullBase>();

            tagsForFrontend = exercise.MetaData.TagWithMetaDatas.Select(p => new TagFullBase()
            {
                Id = p.Tag.Id,
                DisplayName = p.Tag.DisplayName,
                HtmlIcon = p.Tag.HtmlIcon,
                HtmlColor = p.Tag.HtmlColor,
                Description = p.Tag.Description,
                HtmlBackgroundColor = p.Tag.HtmlBackgroundColor,
            }).ToList();


            var tuple = await _context.Exercises
                .Where(p => p.Id == exercise.Id)
                .Select(p => new
                {
                    Releases = p.Releases.Count,
                    Solutions = p.Releases.Sum(r => r.ExerciseReleaseWithUserAsParticipations.Count)
                })
                .AsNoTracking()
                .FirstOrDefaultAsync();

            if (tuple == null)
            {
                await Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "could not get release or solutions count")));
                return;
            }

            var frontendData = new SingleGroupPreviewExerciseTupleFromBackend()
            {
                GroupPreview = new GroupPreviewFromBackend()
                {
                    Id = managingGroupConnection.UserGroup.Id,
                    DisplayName = managingGroupConnection.UserGroup.DisplayName
                },
                GroupRole = new GroupRoleWithPermissionFullBase
                (
                    id: managingGroupConnection.GroupRole.Id,
                    displayName: managingGroupConnection.GroupRole.DisplayName,
                    canChangeExercises: managingGroupConnection.GroupRole.GroupRolePermission.CanChangeExercises,
                    canCreateExercises: managingGroupConnection.GroupRole.GroupRolePermission.CanCreateExercises,
                    canAssessExercises: managingGroupConnection.GroupRole.GroupRolePermission.CanAssessExercises,
                    canDeleteExercises: managingGroupConnection.GroupRole.GroupRolePermission.CanDeleteExercises,
                    canManageExerciseReleases: managingGroupConnection.GroupRole.GroupRolePermission.CanManageExerciseReleases,
                    canAddUserToGroup: managingGroupConnection.GroupRole.GroupRolePermission.CanAddUserToGroup,
                    canLockExercisesPermanently: managingGroupConnection.GroupRole.GroupRolePermission
                        .CanLockExercisesPermanently,
                    canSeeExercisesFromOthersInGroup: managingGroupConnection.GroupRole.GroupRolePermission
                        .CanSeeExercisesFromOthersInGroup,
                    canChangeOtherMembersRole:
                    managingGroupConnection.GroupRole.GroupRolePermission.CanChangeOtherMembersRole,
                    canRemoveMemberFromGroup:
                    managingGroupConnection.GroupRole.GroupRolePermission.CanRemoveMemberFromGroup,
                    canSeeOtherMembers: managingGroupConnection.GroupRole.GroupRolePermission.CanSeeOtherMembers,
                    canChangeGroupData: managingGroupConnection.GroupRole.GroupRolePermission.CanChangeGroupData
                ),
                ExercisePreview = new EditableExercisePreviewFromBackend()
                {
                    Id = exercise.Id,
                    DisplayName = exercise.DisplayName,
                    ShortDescription = exercise.ShortDescription,
                    Tags = tagsForFrontend,
                    LastUpdatedAt = exercise.LastUpdatedAt,
                    Solutions = tuple.Solutions,
                    Releases = tuple.Releases,
                    IsOnlyVisibleToMe = exercise.IsOnlyVisibleToOwner,
                    ManagingGroupId = managingGroupConnection.UserGroup.Id,
                    SupportedPLangIds = exercise.CodeTemplates.Select(p => p.PLangId).ToList(),
                    IsPermanentlyLocked = exercise.IsPermanentlyLocked,

                    Owner = exercise.UserId != null
                        ? new ExerciseOwner(exercise.User.Id, exercise.User.Token, exercise.User.FirstName,
                            exercise.User.LastName) //exercise.User is fully loaded before
                        : null,
                    OwnerId = exercise.UserId
                }
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<SingleGroupPreviewExerciseTupleFromBackend>(ResponseCode.Ok,
                        "own exercises queried",
                        frontendData)));
        }

        /// <summary>
        /// delete the requested exercise
        /// check first that all releases are deleted!!
        /// </summary>
        /// <param name="exerciseId">the exercise id</param>
        [HttpDelete("delete/{exerciseId}")]
        public async Task DeleteExercise(int exerciseId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var tuple = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                //.Include(p => p.UserGroupId) //also get the connected user group that manages the exercise 
                .Select(p => new
                {
                    Exercise = p,
                    ReleaseCount = p.Releases.Count
                }).FirstOrDefaultAsync();

            //check if permission ... or owner

            if (tuple?.Exercise == null) //tuple == null || tuple.Exercise == null
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            var hasPermission =
                await base.HasGroupPermission(tuple.Exercise.UserGroupId,
                    permission => permission != null && permission.CanDeleteExercises);

            //if has permission OR user is the owner then he/she is allowed to delete...
            // --> not allowed if :: no permission AND user is not the owner
            if (!hasPermission && tuple.Exercise.UserId != userId)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no permission")));
                return;
            }

            if (tuple.Exercise.IsPermanentlyLocked)
            {
                await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                    "exercise is permanently locked")));
                return;
            }


            if (tuple.ReleaseCount != 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "cannot delete exercise because it still has releases")));
                return;
            }


            try
            {
                _context.Exercises.Remove(tuple.Exercise);
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "exercise deleted")));
        }

        private IQueryable<ExerciseRelease> _applyTagFilters(IQueryable<ExerciseRelease> query,
            PaginationPostDataWithTagsFilter pagination)
        {
            if (possibleLogicalConnectors.Contains(pagination.LogicalConnector) == false)
            {
                pagination.LogicalConnector = possibleLogicalConnectors.First();
            }

            //e.g. we have the tags 1,2,3 ...
            if (pagination.LogicalConnector == "and")
            {
                //all filter tag needs to be in the exercise tags (and filtered)
                // 1 && 2 --> all where 1 & 2 is present
                if (pagination.SelectedFilterTagIds.Count > 0)
                {
                    //optional check
                    query = query
                            .Where(p =>
                                    pagination.SelectedFilterTagIds.All(tagId =>
                                        p.Exercise.MetaData.TagWithMetaDatas.Any(k =>
                                            k.TagId ==
                                            tagId)) //any is working (e.g. no tags) because all gives true even for emtpy sequences
                            )
                        ;
                }

                //the selectedNegativeFilterTagsIds must not be in the tags from the exercise
                //!2 && !1 --> all where 1 & 2 is missing
                if (pagination.SelectedNegativeFilterTagsIds.Count > 0)
                {
                    //optional check
                    query = query
                            .Where(p =>
                                pagination.SelectedNegativeFilterTagsIds.All(tagId =>
                                    p.Exercise.MetaData.TagWithMetaDatas.All(k => k.TagId != tagId))
                            )
                        ;
                }
            }
            else
            {
                //connector is OR

                //only one tag needs to be in the exercise tags
                //1 || 2 --> all with 1 or 2 e.g. all from 2017 or 2018
                if (pagination.SelectedFilterTagIds.Count > 0)
                {
                    //we need to check else any will always return false
                    query = query
                            .Where(p =>
                                pagination.SelectedFilterTagIds.Any(tagId =>
                                    p.Exercise.MetaData.TagWithMetaDatas.Any(k => k.TagId == tagId)) ||
                                pagination.SelectedNegativeFilterTagsIds.Any(tagId =>
                                    p.Exercise.MetaData.TagWithMetaDatas.All(k => k.TagId != tagId))
                            )
                        ;
                }

                //this won't work because the 2 where statements are connected via AND (executed after another)
                //!1 || !2  == not(1 && 2) --> all that have not 1 and 2 together
                // if (pagination.SelectedNegativeFilterTagsIds.Count > 0 || pagination.SelectedNegativeFilterTagsIds.Count > 0) { //we need to check else any will always return false
                //     query = query
                //         .Where(p =>
                //             pagination.SelectedNegativeFilterTagsIds.Any(tagId =>
                //                 p.Exercise.MetaData.TagWithMetaDatas.All(k => k.TagId != tagId))
                //         )
                //     ;
                // }
            }

            return query;
        }

        private IQueryable<ExerciseReleaseWithUserAsParticipation> _applyTagFilters(
            IQueryable<ExerciseReleaseWithUserAsParticipation> query, PaginationPostDataWithTagsFilter pagination)
        {
            if (possibleLogicalConnectors.Contains(pagination.LogicalConnector) == false)
            {
                pagination.LogicalConnector = possibleLogicalConnectors.First();
            }

            //e.g. we have the tags 1,2,3 ...
            if (pagination.LogicalConnector == "and")
            {
                //all filter tag needs to be in the exercise tags (and filtered)
                // 1 && 2 --> all where 1 & 2 is present
                if (pagination.SelectedFilterTagIds.Count > 0)
                {
                    //optional check
                    query = query
                            .Where(p =>
                                pagination.SelectedFilterTagIds.All(tagId =>
                                    p.ExerciseRelease.Exercise.MetaData.TagWithMetaDatas.Any(k => k.TagId == tagId))
                            )
                        ;
                }

                //the selectedNegativeFilterTagsIds must not be in the tags from the exercise
                //!2 && !1 --> all where 1 & 2 is missing
                if (pagination.SelectedNegativeFilterTagsIds.Count > 0)
                {
                    //optional check
                    query = query
                            .Where(p =>
                                pagination.SelectedNegativeFilterTagsIds.All(tagId =>
                                    p.ExerciseRelease.Exercise.MetaData.TagWithMetaDatas.All(k => k.TagId != tagId))
                            )
                        ;
                }
            }
            else
            {
                //connector is OR

                //only one tag needs to be in the exercise tags
                //1 || 2 --> all with 1 or 2 e.g. all from 2017 or 2018
                if (pagination.SelectedFilterTagIds.Count > 0)
                {
                    //we need to check else any will always return false
                    query = query
                            .Where(p =>
                                pagination.SelectedFilterTagIds.Any(tagId =>
                                    p.ExerciseRelease.Exercise.MetaData.TagWithMetaDatas.Any(k => k.TagId == tagId)) ||
                                pagination.SelectedNegativeFilterTagsIds.Any(tagId =>
                                    p.ExerciseRelease.Exercise.MetaData.TagWithMetaDatas.All(k => k.TagId != tagId))
                            )
                        ;
                }

                //this won't work because the 2 where statements are connected via AND (executed after another)
                //!1 || !2  == not(1 && 2) --> all that have not 1 and 2 together
                // if (pagination.SelectedNegativeFilterTagsIds.Count > 0 || pagination.SelectedNegativeFilterTagsIds.Count > 0) { //we need to check else any will always return false
                //     query = query
                //         .Where(p =>
                //             pagination.SelectedNegativeFilterTagsIds.Any(tagId =>
                //                 p.ExerciseRelease.Exercise.MetaData.TagWithMetaDatas.All(k => k.TagId != tagId))
                //         )
                //     ;
                //  }
            }

            return query;
        }

        private IQueryable<Exercise> _applyTagFilters(IQueryable<Exercise> query,
            PaginationPostDataWithTagsFilter pagination)
        {
            if (possibleLogicalConnectors.Contains(pagination.LogicalConnector) == false)
            {
                pagination.LogicalConnector = possibleLogicalConnectors.First();
            }

            //e.g. we have the tags 1,2,3 ...
            if (pagination.LogicalConnector == "and")
            {
                //all filter tag needs to be in the exercise tags (and filtered)
                // 1 && 2 --> all where 1 & 2 is present
                if (pagination.SelectedFilterTagIds.Count > 0)
                {
                    //optional check
                    query = query
                            .Where(p =>
                                pagination.SelectedFilterTagIds.All(tagId =>
                                    p.MetaData.TagWithMetaDatas.Any(k => k.TagId == tagId))
                            )
                        ;
                }


                //the selectedNegativeFilterTagsIds must not be in the tags from the exercise
                //!2 && !1 --> all where 1 & 2 is missing
                if (pagination.SelectedNegativeFilterTagsIds.Count > 0)
                {
                    //optional check
                    query = query
                            .Where(p =>
                                pagination.SelectedNegativeFilterTagsIds.All(tagId =>
                                    p.MetaData.TagWithMetaDatas.All(k => k.TagId != tagId))
                            )
                        ;
                }
            }
            else
            {
                //connector is OR

                //only one tag needs to be in the exercise tags
                //1 || 2 --> all with 1 or 2 e.g. all from 2017 or 2018
                if (pagination.SelectedFilterTagIds.Count > 0 || pagination.SelectedNegativeFilterTagsIds.Count > 0)
                {
                    //we need to check else any will always return false
                    query = query
                            .Where(p =>
                                pagination.SelectedFilterTagIds.Any(tagId =>
                                    p.MetaData.TagWithMetaDatas.Any(k => k.TagId == tagId)) ||
                                pagination.SelectedNegativeFilterTagsIds.Any(tagId =>
                                    p.MetaData.TagWithMetaDatas.All(k => k.TagId != tagId))
                            )
                        ;
                }


                //this won't work because the 2 where statements are connected via AND (executed after another)
                //!1 || !2  == not(1 && 2) --> all that have not 1 and 2 together
                //  if (pagination.SelectedNegativeFilterTagsIds.Count > 0) { //we need to check else any will always return false
                //     query = query
                //         .Where(p =>
                //             pagination.SelectedNegativeFilterTagsIds.Any(tagId =>
                //                 p.MetaData.TagWithMetaDatas.All(k => k.TagId != tagId))
                //         )
                //     ;
                //  }
            }

            return query;
        }
    }


    public class SingleGroupPreviewExerciseTupleFromBackend
    {
        /// <summary>
        /// the group (preview)
        /// </summary>
        public GroupPreviewFromBackend GroupPreview { get; set; }

        /// <summary>
        /// we need the group role for the current user in the group (groupPreview) to determine what he/she is allowed to do
        /// </summary>
        public GroupRoleWithPermissionFullBase GroupRole { get; set; }

        /// <summary>
        /// the exercise previews in the group
        /// </summary>
        public EditableExercisePreviewFromBackend ExercisePreview { get; set; }
    }

    /// <summary>
    /// a group (just name and id)
    /// </summary>
    public class GroupPreviewFromBackend
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }
    }

    /// <summary>
    /// a exercise preview for the sites that displays the available exercises in a group or where the user is the creator
    /// --> no release data needed
    /// </summary>
    public class EditableExercisePreviewFromBackend
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the short description
        /// </summary>
        public string ShortDescription { get; set; }

        /// <summary>
        /// the tags connected to this exercise
        /// </summary>
        public List<TagFullBase> Tags { get; set; }

        /// <summary>
        /// true: only visible to me in the groups exercise overview, false: do not show to others in group ...
        /// </summary>
        public bool IsOnlyVisibleToMe { get; set; }

        /// <summary>
        /// exercise and all dependent entities should be readonly (no edit/no delete)! (recursively / transitively) 
        /// </summary>
        public bool IsPermanentlyLocked { get; set; }

        /// <summary>
        /// when the exercise was last updated
        /// </summary>
        public DateTime LastUpdatedAt { get; set; }

        /// <summary>
        /// the amount of solutions submitted for this exercise
        /// </summary>
        public int Solutions { get; set; }

        /// <summary>
        /// the amount of releases for this exercsie (used to determine of the exercise can be deleted (when releases == 0) or not)
        /// </summary>
        public int Releases { get; set; }

        /// <summary>
        /// the group id that manages the exercise (to get the user permission)
        /// </summary>
        public int ManagingGroupId { get; set; }

        /// <summary>
        /// the plang ids the exercise supports (has code templates for)
        /// </summary>
        public List<int> SupportedPLangIds { get; set; }

        public int? OwnerId { get; set; }

        /// <summary>
        /// the exercise owner OR null if the owner was removed
        /// </summary>
        public ExerciseOwner Owner { get; set; }
    }

    public class ExerciseOwner
    {
        //we cant use id because ef gives error when trying to call this constructor with (exercise.User.Id) (error: could be null?) 
        public ExerciseOwner(int id, string token, string firstName, string lastName)
        {
            Id = id;
            Token = token;
            FirstName = firstName;
            LastName = lastName;
        }

        public int Id { get; set; }
        public string Token { get; set; }

        public string FirstName { get; set; }
        public string LastName { get; set; }
    }

    /// <summary>
    /// a preview for an exercise (just some basic information)
    /// this is mainly used for exercise with release where we attach some user specific information e.g. ShouldNotCount
    /// </summary>
    public class ExercisePreviewFromBackend
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the short description
        /// </summary>
        public string ShortDescription { get; set; }

        /// <summary>
        /// the code from the related exercise release
        /// </summary>
        public string GeneratedCode { get; set; }

        /// <summary>
        /// true: assessment result should not count, false: not
        /// </summary>
        public bool ShouldNotCount { get; set; }

        /// <summary>
        /// exercise and all dependent entities should be readonly (no edit/no delete)! (recursively / transitively) 
        /// </summary>
        public bool IsPermanentlyLocked { get; set; }

        /// <summary>
        /// the plang the release
        /// </summary>
        public int ReleasedForPLangId { get; set; }

        /// <summary>
        /// the tags connected to this exercise
        /// </summary>
        public List<TagFullBase> Tags { get; set; }

        /// <summary>
        /// the release start type (we cannot rely on releaseDurationType to check if the release will end)
        /// if the release start type is automatic then the releaseDurationType should be taken into account
        /// </summary>
        public int ReleaseStartType { get; set; }

        /// <summary>
        /// the duration type (<see cref="AutomaticEndAt"/>) for the release
        /// </summary>
        public int ReleaseDurationType { get; set; }

        /// <summary>
        /// the automatic ending time (if the <see cref="ReleaseDurationType"/> is automatic)
        /// </summary>
        public DateTime? AutomaticEndAt { get; set; }

        /// <summary>
        /// if true apply the <see cref="AvailableWorkingTimeInMinutes"/> else unlimited working time
        /// </summary>
        public bool HasLimitedWorkingTime { get; set; }

        /// <summary>
        /// the time the student can spend at the <see cref="Exercise"/>
        /// </summary>
        public int AvailableWorkingTimeInMinutes { get; set; }

        /// <summary>
        /// true: the exercise is visible to all users event if they not entered the code, false: not
        /// </summary>
        public bool IsVisibleToAllAfterRelease { get; set; }

        /// <summary>
        /// the date time when the user has accessed the exercise for the first time (not actually accessed but could have accessed e.g. wehn he/she enters the release code)
        /// used to calculate the left working time...
        /// can be null because when accessed without code we know nothing about the (future) participation
        /// </summary>
        public DateTime? FirstAccessedAt { get; set; }

        /// <summary>
        /// true: is still released, false: not
        /// </summary>
        public bool IsReleased { get; set; }
    }
}
