﻿using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using ClientServer.Db;
using ClientServer.Models.Exercises.AfterSolutions;
using ClientServer.Models.Exercises.Solution;

namespace ClientServer.Models.Exercises.Tests
{
    /// <summary>
    /// a test case for an <see cref="Exercise"/>
    /// </summary>
    public class Test
    {

        public Test()
        {
            TestAssets = new List<TestAsset>();
            AssetReferences = new List<TestWithFileAsAssetReference>();
            TestResultsNew = new List<TestWithSingleSolutionAsTestResult>();
            AfterTestResults = new List<TestWithAfterSolutionAsTestResult>();
        }

        public int Id { get; set; }
        /// <summary>
        /// the display name
        /// </summary>
        [MaxLength(YapexDbContext.DefaultMaxStringLength)]
        public string DisplayName { get; set; }
        /// <summary>
        /// true: this <see cref="Test"/> is a submit test (is will run after the deadline is completed),
        /// <para></para>
        /// false: this is a normal test which could be rerun
        /// </summary>
        public bool IsSubmitTest { get; set; }
        /// <summary>
        /// the content for the <see cref="Test"/> ... see the black-box-test protocol
        /// </summary>
        [MaxLength(YapexDbContext.TestContentMaxStringLength)]
        public string Content { get; set; }
        /// <summary>
        /// the weight of the <see cref="Test"/>
        /// </summary>
        public int Weight { get; set; }

        /// <summary>
        /// the index for the frontend sorting (>= 0, 0 is first, 1 second, ...)
        /// </summary>
        public int DisplayIndex { get; set; }


        //foreign

        /// <summary>
        /// all related assets for this test e.g. a file to read
        /// </summary>
        public List<TestAsset> TestAssets { get; set; }
        
        /// <summary>
        /// all related assets for this test e.g. a file to read
        /// </summary>
        public List<TestWithFileAsAssetReference> AssetReferences { get; set; }


        public int TestSettingsId { get; set; }
        /// <summary>
        /// the connected <see cref="TestSettings"/>
        /// </summary>
        public TestSettings TestSettings { get; set; }


        public int TestTypeId { get; set; }
        /// <summary>
        /// the connected <see cref="TestType"/>
        /// </summary>
        public TestType TestType { get; set; }


        public int ExerciseId { get; set; }
        /// <summary>
        /// the connected <see cref="Exercise"/>
        /// </summary>
        public Exercise Exercise { get; set; }
        
        //---added
        
        public List<TestWithSingleSolutionAsTestResult> TestResultsNew { get; set; }
        
        public List<TestWithAfterSolutionAsTestResult> AfterTestResults { get; set; }
    }
}
